
from urllib.parse import urlparse, urlunparse


def noslash(txt: str) -> str:
    """Remove leading and trailing slashes from the given text.

    Args:
        txt (str): The text.

    Returns:
        str: The text without leading and trailing slashes.
    """    
    return txt.removeprefix("/").removesuffix("/")


def endslash(txt: str) -> str:
    """Add a trailing slash to the given text if not present.

    Args:
        txt (str): The text.

    Returns:
        str: The text with a trailing slash.
    """    
    return txt.removesuffix("/") + "/"


def http_href_to_s3(href: str) -> str:
    """Generate an S3 href from an HTTP href

    Args:
        href (str): The HTTP href.

    Returns:
        str: The S3 href.
    """ 
    url = urlparse(href)
    components = list(url[:])
    if len(components) == 5:
        components.append('')
    components[0] = "s3"
    components[1] = ""
    components[2] = components[2].removeprefix("/")
    s3url = urlunparse(tuple(components))
    return s3url


def join_pathes(*ps: str) -> str:
    """Join multiple pathes ensuring there is exactly one '/' between them.

    Returns:
        str: The joined path.
    """    
    pathes: list[str] = list(filter(lambda p: p != "", ps))
    if len(pathes) == 1:
        return pathes[0]
    if len(pathes) > 1:
        pathes[0] = pathes[0].removesuffix("/")
        pathes[-1] = pathes[-1].removeprefix("/")
        for i in range(1, len(pathes) - 1):
            pathes[i] = pathes[i].removeprefix("/").removesuffix("/")
    return "/".join(pathes)



import tempfile
from typing import Literal
from typing import Annotated, Union
from pydantic import BaseModel, Field, computed_field
import enum
import json


class AccessType(enum.Enum):
    READ = "read"
    WRITE = "write"


class StorageConfiguration(BaseModel, extra='allow'):
    type: str = Field(title='Storage type', description="Type of the storage used")
    is_local: bool = Field(title='Is a local storage', description="Whether the storage is local or remote")
    writable_paths: list[str] = Field(default=[], title="Writable Paths", description="List of paths where files can be written")
    readable_paths: list[str] = Field(default=[], title="Readable Paths", description="List of paths from which files can be read")


class FileStorageConfiguration(StorageConfiguration):
    type: Literal["file"] = Field(default="file", title="Type", description="Indicates the storage type, fixed to 'file'")
    is_local: Literal[True] = Field(default=True, title='Is a local storage', description="Whether the storage is local or remote")


class GoogleStorageConstants(str, enum.Enum):
    AUTH_URI = "https://accounts.google.com/o/oauth2/auth"
    TOKEN_URI = "https://oauth2.googleapis.com/token"
    AUTH_PROVIDER_CERT_URL = "https://www.googleapis.com/oauth2/v1/certs"
    UNIVERSE_DOMAIN = "googleapis.com"


class GoogleStorageApiKey(BaseModel):
    type: Literal["service_account"] = Field(default="service_account", title="Type", description="Must be 'service_account'.")
    project_id: str = Field(title="Project ID", description="Google Cloud project identifier")
    private_key_id: str = Field(title="Private Key ID", description="ID of the private key used for authentication")
    private_key: str = Field(title="Private Key", description="The private key content in PEM format")
    client_email: str = Field(title="Client Email", description="Service account email address")
    client_id: str | None = Field(default=None, title="Client ID", description="Optional client ID of the service account")
    auth_uri: Literal[GoogleStorageConstants.AUTH_URI] = Field(default=GoogleStorageConstants.AUTH_URI.value,
        title="Auth URI", description="OAuth2 auth endpoint URI")
    token_uri: Literal[GoogleStorageConstants.TOKEN_URI] = Field(default=GoogleStorageConstants.TOKEN_URI.value,
        title="Token URI", description="OAuth2 token endpoint URI")
    auth_provider_x509_cert_url: Literal[GoogleStorageConstants.AUTH_PROVIDER_CERT_URL] = Field(
        default=GoogleStorageConstants.AUTH_PROVIDER_CERT_URL.value,
        title="Provider X.509 Cert URL", description="URL for the provider's X.509 certificate")
    universe_domain: Literal[GoogleStorageConstants.UNIVERSE_DOMAIN] = Field(default=GoogleStorageConstants.UNIVERSE_DOMAIN.value,
        title="Universe Domain", description="Domain of the target universe (typically 'googleapis.com')")

    @computed_field
    @property
    def client_x509_cert_url(self) -> str:
        return f"https://www.googleapis.com/robot/v1/metadata/x509/{self.client_email.replace('@', '%40')}"


class GoogleStorageConfiguration(StorageConfiguration):
    type: Literal["gs"] = Field(default="gs", title="Type", description="Indicates the storage type, fixed to 'gs'")
    is_local: Literal[False] = Field(default=False, title='Is a local storage', description="Whether the storage is local or remote")
    bucket: str = Field(title="Bucket name", description="Name of the Google Cloud Storage bucket")
    api_key: GoogleStorageApiKey | None = Field(title="API Key", description="API key for storage authentication", default=None)

    @computed_field
    @property
    def is_anon_client(self) -> bool:
        return self.api_key is None

    @computed_field
    @property
    def credentials_file(self) -> str:
        if not self.is_anon_client:
            with tempfile.NamedTemporaryFile("w+", delete=False) as f:
                json.dump(self.api_key.model_dump(exclude_none=True, exclude_unset=True), f)
                f.close()
            credentials = f.name
        else:
            credentials = None
        return credentials


class HttpStorageConfiguration(StorageConfiguration):
    type: Literal["http"] = Field(default="http", title="Type", description="Indicates the storage type, fixed to 'http'")
    is_local: Literal[False] = Field(default=False, title='Is a local storage', description="Whether the storage is local or remote")
    headers: dict[str, str] = Field(default={}, title="Headers", description="Additional HTTP headers to include in each request")
    domain: str = Field(title="Domain", description="Domain used for HTTP storage endpoint, e.g. 'example.com'")
    force_download: bool = Field(default=False, title="Force Download", description="If true, always download the file instead of caching."
    )


class S3ApiKey(BaseModel):
    access_key: str = Field(title="S3 Access API key", description="Access api key for S3 storage authentication")
    secret_key: str = Field(title="S3 Secret API key", description="Secret api key for S3 storage authentication")


class S3StorageConfiguration(StorageConfiguration):
    type: Literal["s3"] = Field(default="s3", title="Type", description="Indicates the storage type, fixed to 's3'")
    is_local: Literal[False] = Field(default=False, title='Is a local storage', description="Whether the storage is local or remote")
    bucket: str = Field(title="Bucket name", description="Name of the S3 bucket")
    region: str = Field(default="auto", title="Region", description="Region of the bucket")
    endpoint: str = Field(title="Endpoint", description="Endpoint to access S3 storage")
    api_key: S3ApiKey | None = Field(title="API Key", description="API key for storage authentication", default=None)
    max_objects: int = Field(default=1000, title="Max object", description="Maximum number of objects to fetch when listing elements in a directory")

    @computed_field
    @property
    def is_anon_client(self) -> bool:
        return self.api_key is None


class HttpsStorageConfiguration(HttpStorageConfiguration):
    type: Literal["https"] = "https"


AnyStorageConfiguration = Annotated[Union[FileStorageConfiguration, GoogleStorageConfiguration, HttpStorageConfiguration, HttpsStorageConfiguration, S3StorageConfiguration], Field(discriminator="type")]


class AccessManagerSettings(BaseModel):
    storages: list[AnyStorageConfiguration] = Field(title="Storage list", description="List of configurations for the available storages")
    tmp_dir: str = Field(title="Temporary directory", description="Temporary directory in which to write files that will be deleted")



import os
import shutil
from contextlib import contextmanager
from typing import Annotated, Union
from urllib.parse import urlparse

from pydantic import Field

from aias_common.access.configuration import AccessManagerSettings
from aias_common.access.file import File
from aias_common.access.logger import Logger
from aias_common.access.storages.file import AccessType, FileStorage
from aias_common.access.storages.gs import GoogleStorage
from aias_common.access.storages.http import HttpStorage
from aias_common.access.storages.https import HttpsStorage
from aias_common.access.storages.s3 import S3Storage

AnyStorage = Annotated[Union[FileStorage, GoogleStorage, HttpStorage, HttpsStorage, S3Storage], Field(discriminator="type")]

LOGGER = Logger.logger


class AccessManager:
    storages: list[AnyStorage]
    tmp_dir: str
    cache_ttl: int = 60 * 60 * 24
    cache_size = 1024

    @staticmethod
    def init(ams: AccessManagerSettings):
        LOGGER.info("Initializing storages Access Manager")
        AccessManager.storages = []
        for s in ams.storages:
            match s.type:
                case "file":
                    AccessManager.storages.append(FileStorage(s))
                case "gs":
                    AccessManager.storages.append(GoogleStorage(s))
                case "http":
                    AccessManager.storages.append(HttpStorage(s))
                case "https":
                    AccessManager.storages.append(HttpsStorage(s))
                case "s3":
                    AccessManager.storages.append(S3Storage(s))
                case _:
                    raise NotImplementedError(f"Specified storage {s.storage_configuration.type} is not implemented")

        tmp_dir = ams.tmp_dir
        is_tmp_dir_authorized = any(
            map(lambda s: s.is_path_authorized(tmp_dir, AccessType.WRITE),
                filter(lambda s: s.storage_configuration.type == "file", AccessManager.storages)))
        if not is_tmp_dir_authorized:
            raise PermissionError("The given tmp_dir is not part of any defined FileStorage with WRITE authorization")

        AccessManager.tmp_dir = tmp_dir

    @staticmethod
    def resolve_storage(href: str) -> AnyStorage:
        """
        Based on the defined storages, returns the one matching the input href
        """

        for s in AccessManager.storages:
            try:
                if s.supports(href):
                    return s
            except Exception:
                ...
        raise PermissionError(f"Storage for {href} is not configured")

    @staticmethod
    def get_local_storage() -> FileStorage:
        """
        Based on the defined storages, returns the one that is a file storage
        """

        for s in AccessManager.storages:
            if s.get_configuration().is_local:
                return s
        raise Exception("No local storage configured")

    @staticmethod
    def get_storage_parameters(href: str) -> dict:
        storage = AccessManager.resolve_storage(href)

        return storage.get_storage_parameters()

    @staticmethod
    def check_path_writable(href: str):
        """
        Checks that the path is a writable path for at least one of the file storages
        """
        if not AccessManager.resolve_storage(href=href).is_path_authorized(href, AccessType.WRITE):
            raise PermissionError(f"Path '{href}' is not writable")

    @staticmethod
    def check_path_readable(href: str):
        """
        Checks that the path is a readable path for at least one of the file storages
        """
        if not AccessManager.resolve_storage(href=href).is_path_authorized(href, AccessType.READ):
            raise PermissionError(f"Path '{href}' is not readable")

    @staticmethod
    def push(href: str, dst: str):
        """
        Push a file from a local storage to write it in a storage.
        If the destination storage is local, then it is a copy. Otherwise it is an upload.
        """
        # Check that href is local
        scheme = urlparse(href).scheme
        if scheme != "" and scheme != "file":
            raise ValueError("Source file to upload must be on the local filesystem")

        storage = AccessManager.resolve_storage(dst)
        storage.push(href, dst)

    @staticmethod
    def pull(href: str, dst: str):
        """
        Pulls a file from a storage to write it in the local storage.
        If the input storage is local, then it is a copy. Otherwise it is a download.
        """
        # Check that dst is local
        scheme = urlparse(dst).scheme
        if scheme != "" and scheme != "file":
            raise ValueError("Destination must be on the local filesystem")

        storage = AccessManager.resolve_storage(href)
        storage.pull(href, dst)

    @staticmethod
    @contextmanager
    def stream(href: str):
        """
        Reads the content of a file in a storage without downloading it.
        """
        with AccessManager.resolve_storage(href=href).stream(href) as f:
            yield f

    @staticmethod
    def get_rasterio_session(href: str):
        storage = AccessManager.resolve_storage(href)
        return storage.get_rasterio_session(href)

    @staticmethod
    def exists(href: str) -> bool:
        """
        Whether the file exists
        """
        storage = AccessManager.resolve_storage(href)
        return storage.exists(href)

    @staticmethod
    def is_download_required(href: str):
        storage = AccessManager.resolve_storage(href)

        return storage.storage_configuration.type in ["http", "https"] \
            and storage.get_configuration().force_download

    @staticmethod
    @contextmanager
    def make_local(href: str, dst: str | None = None):
        """Prepare the file to be processed locally. Once the file has been used, if it has been pulled, deletes it.

        Args:
            href (str): Href (local or not) of the file

        Returns:
            str: The local path at which the file can be found
        """
        storage = AccessManager.resolve_storage(href)

        # If the storage is not local, pull it
        if not storage.storage_configuration.is_local:
            if dst is None:
                dst = os.path.join(AccessManager.tmp_dir, os.path.basename(href))

            AccessManager.pull(href, dst)
            try:
                yield dst
            finally:
                AccessManager.clean(dst)  # !DELETE!
        else:
            yield href

    @staticmethod
    @contextmanager
    def make_local_list(href_list: list[str], dst_list: list[str | None] | None = None):
        """Prepare a list of files to make them available locally for further processing.
           Once used, the file is deleted if it has been pulled
        """
        # Check that the input lists match each other length
        if dst_list is not None and len(href_list) != len(dst_list):
            raise ValueError("Input href and dst must have the same length")
        if dst_list is None:
            dst_list = [None for _ in href_list]

        # For each of the input pair of (href, dst), check if the corresponding storage is local
        # If not, pull it, store dst and tag the iteration as pulled. Otherwise, store href and tag as not pulled.
        # Once all local, will yield the list of local paths
        # Cleanup will only remove the files that were pulled
        local_href_list: list[str] = []
        was_pulled = []
        try:
            for href, dst in zip(href_list, dst_list):
                storage = AccessManager.resolve_storage(href)

                if not storage.storage_configuration.is_local:
                    if dst is None:
                        dst = os.path.join(AccessManager.tmp_dir, os.path.basename(href))

                    AccessManager.pull(href, dst)
                    local_href_list.append(dst)
                    was_pulled.append(True)
                else:
                    local_href_list.append(href)
                    was_pulled.append(False)
            yield local_href_list
        finally:
            for pulled, local_href in zip(was_pulled, local_href_list):
                # Only pulled files (files downloaded by this process) are deleted, as a clean up procedure.
                if pulled:
                    AccessManager.clean(local_href)  # !DELETE!

    @staticmethod
    def clean(href: str):
        storage = AccessManager.resolve_storage(href)
        storage.clean(href)  # !DELETE!

    @staticmethod
    def zip(href: str, zip_path: str):
        with AccessManager.make_local(href) as local_href:
            # Get direct parent folder of href_file to zip
            dir_name = os.path.dirname(local_href)
            shutil.make_archive(zip_path, 'zip', dir_name)

    @staticmethod
    def is_file(href: str):
        storage = AccessManager.resolve_storage(href)

        return storage.is_file(href)

    @staticmethod
    def is_dir(href: str):
        storage = AccessManager.resolve_storage(href)

        return storage.is_dir(href)

    @staticmethod
    def get_size(href: str):
        storage = AccessManager.resolve_storage(href)
        if AccessManager.exists(href):
            if AccessManager.is_file(href):
                return storage.get_file_size(href)
            else:
                folder_size = 0
                for f in AccessManager.listdir(href):
                    folder_size += AccessManager.get_size(f.path)
                return folder_size
        raise FileNotFoundError(f"Given href does not exist: {href}")

    @staticmethod
    def listdir(href: str) -> list[File]:
        storage = AccessManager.resolve_storage(href)

        if not storage.is_dir(href):
            raise NotADirectoryError(f"Given href does not point to a directory: {href}")

        return storage.listdir(href)

    @staticmethod
    def get_last_modification_time(href: str) -> float | None:
        storage = AccessManager.resolve_storage(href)
        return storage.get_last_modification_time(href)

    @staticmethod
    def get_creation_time(href: str) -> float | None:
        storage = AccessManager.resolve_storage(href)
        return storage.get_creation_time(href)

    @staticmethod
    def makedir(href: str, strict=False):
        """
        Create if needed (and possible) the specified dir
        """
        storage = AccessManager.resolve_storage(href)
        return storage.makedir(href, strict)

    @staticmethod
    def dirname(href: str):
        """
        Wraps os.path.dirname to allow for absolute path to be determined if needed
        """
        storage = AccessManager.resolve_storage(href)
        return storage.dirname(href)

    @staticmethod
    def get_gdal_src(href: str):
        """
        Returns the dataset of an archive through GDAL without pulling the file
        """
        storage = AccessManager.resolve_storage(href)
        return storage.get_gdal_src(href)

    @staticmethod
    def get_gdal_md(href: str):
        """
        Returns the metadata of an archive through GDAL without pulling the file
        """
        return AccessManager.get_gdal_src(href).GetMetadata()

    @staticmethod
    def get_gdal_proj(href: str):
        """
        Returns the projection of an archive through GDAL without pulling the file
        """
        return AccessManager.get_gdal_src(href).GetProjection()

    @staticmethod
    def get_gdal_info(href: str, gdal_options):
        """
        Returns the info of an archive through GDAL without pulling the file
        """
        storage = AccessManager.resolve_storage(href)
        return storage.get_gdal_info(href, gdal_options)



from fastapi import APIRouter, status
from fastapi.responses import Response

ROUTER = APIRouter()


@ROUTER.get("/healthcheck")
async def healthcheck():
    return Response(content="OK", status_code=status.HTTP_200_OK)
